// ---------------------------------
// Filename      : GeneralTools.cpp
// Author        : Sven Maerivoet
// Last modified : 25/08/2020
// Target        : C++ (Arduino UNO)
//
// All rights reserved.
// ---------------------------------

/*
  Copyright 2017, 2020 Sven Maerivoet

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
*/

#include <GeneralTools.h>

void accelerateADCMeasurements()
{
	// set ADC prescale to 16 to have a clock of 1 MHz and a sample rate of 76.8 KHz
	sbi(ADCSRA,ADPS2);
	cbi(ADCSRA,ADPS1);
	cbi(ADCSRA,ADPS0);
}

void setRandomSeed(int analogPin)
{
	// take a random reading from an analog pin to seed the Arduino's pseudo-RNG
	randomSeed(analogRead(analogPin));
}

void setRandomSeed()
{
	// take a random reading from analog pin A5 to seed the Arduino's pseudo-RNG
	setRandomSeed(A5);
}

/*
byte fPWMValues[256];
void createPWMLookupTable()
{
	// there are 256 PWM steps, so we use a 256-base logarithm to convert them to [0,1] and rescale them then again to [0,255]
	for (int pwm = 0; pwm < 256; ++pwm) {
		fPWMValues[pwm] = floor((log((double) pwm + 1.0) / log(256.0)) * 255.0);
	}
}
*/

// converts a percentage between 0.0% and 100.0% to the [0,255] range
byte getNonLinearPWMPercentageBased(double percentage)
{
	// approximate the non-linear curve
	double a = 9.7758463166360387E-01;
	double b = 5.5498961535023345E-02;
	return (byte) (floor((a * exp(b * percentage) + 0.5)) - 1.0);
}

// converts a value from the [0,255] range to the same range
byte getNonLinearPWM8BitBased(byte value)
{
	return getNonLinearPWMPercentageBased(((double) value / 255.0) * 100.0);
}

/*
 * https://playground.arduino.cc/Code/PwmFrequency/
 * Divides a given PWM pin frequency by a divisor.
 *
 * The resulting frequency is equal to the base frequency divided by
 * the given divisor:
 *   - Base frequencies:
 *      o The base frequency for pins 3, 9, 10, and 11 is 31250 Hz.
 *      o The base frequency for pins 5 and 6 is 62500 Hz.
 *   - Divisors:
 *      o The divisors available on pins 5, 6, 9 and 10 are:
 *          1, 8, 64, 256, 1024
 *      o The divisors available on pins 3 and 11 are:
 *          1    (31250 Hz)
 *          8    (3906 Hz)
 *          64   (488 Hz)
 *          128  (244 Hz)
 *          256  (122 Hz)
 *          1024 (30 Hz)
 *
 * PWM frequencies are tied together in pairs of pins. If one in a
 * pair is changed, the other is also changed to match:
 *   - Pins 5 and 6 are paired on timer0
 *   - Pins 9 and 10 are paired on timer1
 *   - Pins 3 and 11 are paired on timer2
 *
 * Note that this function will have side effects on anything else
 * that uses timers:
 *   - Changes on pins 3, 5, 6, or 11 may cause the delay() and
 *     millis() functions to stop working. Other timing-related
 *     functions may also be affected.
 *   - Changes on pins 9 or 10 will cause the Servo library to function
 *     incorrectly.
 *
 * Thanks to macegr of the Arduino forums for his documentation of the
 * PWM frequency divisors. His post can be viewed at:
 * https://forum.arduino.cc/index.php?topic=16612#msg121031
*/
void setPWMFrequency(int pin, int divisor)
{
	byte mode;
	if ((pin == 5) || (pin == 6) || (pin == 9) || (pin == 10)) {
		switch(divisor) {
			case 1: mode = 0x01; break;
			case 8: mode = 0x02; break;
			case 64: mode = 0x03; break;
			case 256: mode = 0x04; break;
			case 1024: mode = 0x05; break;
			default: return;
		}
		if ((pin == 5) || (pin == 6)) {
			TCCR0B = TCCR0B & 0b11111000 | mode;
		}
		else {
			TCCR1B = TCCR1B & 0b11111000 | mode;
		}
	}
	else if ((pin == 3) || (pin == 11)) {
		switch(divisor) {
			case 1: mode = 0x01; break;
			case 8: mode = 0x02; break;
			case 32: mode = 0x03; break;
			case 64: mode = 0x04; break;
			case 128: mode = 0x05; break;
			case 256: mode = 0x06; break;
			case 1024: mode = 0x07; break;
			default: return;
		}
		TCCR2B = TCCR2B & 0b11111000 | mode;
	}
}

long getVcc(AbstractToneGenerator* toneGenerator)
{
	long result;

	// read 1.1V reference against AVcc
	ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);

	// wait for Vref to settle
	if (toneGenerator != NULL) {
		toneGenerator->generateDelay(2);
	}
	else {
		delay(2);
	}
	ADCSRA |= _BV(ADSC);

	// convert
	while (bit_is_set(ADCSRA,ADSC));

	result = ADCL;
	result |= ADCH<<8;

	// back-calculate AVcc in mV
	result = 1125300L / result;

	return result;
}

long getVcc()
{
	return getVcc(NULL);
}

void serialPrintIntegerVar(String description, int value)
{
	Serial.print(description);
	Serial.print(" -> ");
	Serial.println(value);
}

void serialPrintDoubleVar(String description, double value)
{
	Serial.print(description);
	Serial.print(" -> ");
	Serial.println(value);
}

void serialPrintBooleanVar(String description, bool value)
{
	Serial.print(description);
	Serial.print(" -> ");
	Serial.println(value);
}
